import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join("..", "..", "src")))

from cappr import __version__

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "CAPPr"
copyright = "2023, kddubey"
author = "kddubey"

version = __version__
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_copybutton",
    "sphinx_togglebutton",
]
templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_css_files = [os.path.join("css", "custom.css")]
html_context = {
    "display_github": True,
    "github_user": "kddubey",
    "github_repo": "cappr",
    "github_version": "main",
    "doc_path": "/docs/source/",
}
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# see example: https://github.com/pydata/pydata-sphinx-theme/blob/185a37aa36820f77bffa4c87a772092e9e7cc380/docs/conf.py#L116C12-L116C12
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/kddubey/cappr",
            "icon": "fa-brands fa-github",
        },
    ]
}


def setup(self):
    # Use cappr.Example not cappr.example.Example
    from cappr import Example

    Example.__module__ = "cappr"


def linkcode_resolve(domain, info):
    import importlib
    import inspect

    code_url = "https://github.com/kddubey/cappr/blob/main"
    ## ty https://github.com/readthedocs/sphinx-autoapi/issues/202#issuecomment-907582382
    if domain != "py":
        return
    if not info["module"]:
        return

    mod = importlib.import_module(info["module"])

    if info["fullname"] == "Example":
        Example = getattr(mod, info["fullname"])
        Example.__module__ = "cappr._example"

    if "." in info["fullname"]:
        objname, attrname = info["fullname"].split(".")
        obj = getattr(mod, objname)
        try:
            # object is a method of a class
            obj = getattr(obj, attrname)
        except AttributeError:
            # object is an attribute of a class
            return None
    else:
        obj = getattr(mod, info["fullname"])

    ## Unwrap the object to get the correct source file in case that is wrapped by a
    ## decorator
    obj = inspect.unwrap(obj)

    try:
        file = inspect.getsourcefile(obj)
        lines = inspect.getsourcelines(obj)
    except TypeError:
        # e.g. object is a typing.Union
        return None
    file = os.path.relpath(file, os.path.abspath(".."))
    start, end = lines[1], lines[1] + len(lines[0]) - 1

    if info["fullname"] == "Example":
        Example = getattr(mod, info["fullname"])
        Example.__module__ = "cappr"

    file = file.lstrip("../")
    return f"{code_url}/{file}#L{start}-L{end}"


# ty https://stackoverflow.com/q/67473396/18758987
autodoc_type_aliases = {
    "Sequence": "Sequence",
}



"""
Generic implementation for non-cached LM interfaces
"""

from __future__ import annotations
from functools import lru_cache
from typing import Any, Callable, cast, Literal, Sequence

from cappr.utils import _batch, _check


@lru_cache()
def _does_tokenizer_need_prepended_space(
    tokenize: Callable[[Sequence[str]], list[list[int]]], bos_token_id: int | None
):
    tokenize_single_text = lambda text: tokenize([text])[0]
    return _check.does_tokenizer_need_prepended_space(
        tokenize_single_text, bos_token_id
    )


def _slice_completions(
    completions: Sequence[str],
    end_of_prompt: str,
    log_probs: Sequence[Sequence[float]],
    tokenize: Callable[[Sequence[str]], list[list[int]]],
    bos_token_id: int | None,
) -> list[list[float]]:
    """
    Slice the completion's tokens from each list of log-probabilities in `log_probs`.
    """
    if len(completions) != len(log_probs):
        raise ValueError(
            f"Different numbers of completions and log_probs: {len(completions)}, "
            f"{len(log_probs)}, likely due to an issue with the token_logprobs function"
        )  # pragma: no cover
    if not _does_tokenizer_need_prepended_space(tokenize, bos_token_id):
        end_of_prompt = ""
    completions = [end_of_prompt + completion for completion in completions]
    completion_lengths = [len(tokens) for tokens in tokenize(completions)]
    return [
        log_probs_text[-num_completion_tokens:]
        for num_completion_tokens, log_probs_text in zip(completion_lengths, log_probs)
    ]


def log_probs_conditional(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    end_of_prompt: Literal[" ", ""],
    token_logprobs: Callable[[Sequence[str], Any], list[list[float]]],
    tokenize_completions: Callable[[list[str]], list[list[int]]],
    bos_token_id: int | None,
    *token_logprobs_args,
    **token_logprobs_kwargs,
) -> list[list[list[float]]]:
    texts = [
        prompt + end_of_prompt + completion
        for prompt in prompts
        for completion in completions
    ]
    log_probs_texts = token_logprobs(
        texts,
        *token_logprobs_args,
        end_of_prompt="",
        **token_logprobs_kwargs,
    )
    # log_probs_texts is a 2-D list. Batch it by the size and order of completions to
    # fulfill the spec
    return [
        _slice_completions(
            completions, end_of_prompt, log_probs, tokenize_completions, bos_token_id
        )
        for log_probs in _batch.constant(log_probs_texts, size=len(completions))
    ]


def log_probs_conditional_examples(
    examples,
    token_logprobs: Callable[[Sequence[str], Any], list[list[float]]],
    tokenize_completions: Callable[[Sequence[str]], list[list[int]]],
    bos_token_id: int | None,
    *token_logprobs_args,
    **token_logprobs_kwargs,
) -> list[list[list[float]]]:
    from cappr import Example

    # examples is always a Sequence[Example] b/c of the decorator
    examples = cast(Sequence[Example], examples)

    texts = [
        example.prompt + example.end_of_prompt + completion
        for example in examples
        for completion in example.completions
    ]
    log_probs_texts = token_logprobs(
        texts, *token_logprobs_args, end_of_prompt="", **token_logprobs_kwargs
    )
    should_end_of_prompt_be_empty = not _does_tokenizer_need_prepended_space(
        tokenize_completions, bos_token_id
    )
    end_of_prompts = [
        "" if should_end_of_prompt_be_empty else example.end_of_prompt
        for example in examples
    ]
    completions = [
        end_of_prompt + completion
        for end_of_prompt, example in zip(end_of_prompts, examples)
        for completion in example.completions
    ]
    end_of_prompt = ""  # we already added it in completions
    log_probs_completions = _slice_completions(
        completions, end_of_prompt, log_probs_texts, tokenize_completions, bos_token_id
    )
    # log_probs_completions is a 2-D list. Batch it by the size and order of completions
    # to fulfill the spec
    num_completions_per_prompt = [len(example.completions) for example in examples]
    return list(
        _batch.variable(log_probs_completions, sizes=num_completions_per_prompt)
    )



"""
Unit and integration tests for `cappr.huggingface.classify`. Works by checking that its
functions' outputs are numerically close to those from
`cappr.huggingface.classify_no_cache`.
"""

from __future__ import annotations
from contextlib import nullcontext
import os
import sys
from typing import Sequence

import pytest

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, PreTrainedTokenizerBase
from transformers.modeling_outputs import CausalLMOutput
import huggingface_hub as hf_hub

from cappr import Example
from cappr.huggingface import classify, classify_no_cache
from cappr import huggingface as hf
from cappr.huggingface._utils import BatchEncodingPT, ModelForCausalLM

# sys hack to import from parent
sys.path.insert(1, os.path.join(sys.path[0], ".."))
import _base
import _test_content
from _protocol import classify_module


########################################################################################
###################################### Fixtures ########################################
########################################################################################


@pytest.fixture(
    scope="module",
    params=[
        "hf-internal-testing/tiny-random-GPT2LMHeadModel",
        "HuggingFaceH4/tiny-random-LlamaForCausalLM",
        # These models are reduntant. They're all transformers w/ BPE or SentencePiece
        # tokenization and a non-None eos_token
        # "hf-internal-testing/tiny-random-GPTJForCausalLM",
        # "hf-internal-testing/tiny-random-GPTNeoXForCausalLM",
        # "hf-internal-testing/tiny-random-MistralForCausalLM",
    ],
)
def model_name(request: pytest.FixtureRequest) -> str:
    return request.param


@pytest.fixture(scope="module")
def model(model_name: str) -> ModelForCausalLM:
    model: ModelForCausalLM = AutoModelForCausalLM.from_pretrained(model_name)
    # Set attributes to values that would break CAPPr, if not for the context managers
    model.train()  # LMs w/ dropout (GPT-2) will cause mismatched logits b/c random
    model.config.return_dict = False  # out.logits fails (not for transformers>=4.31)
    setattr(model.config, "use_cache", False)  # out.past_key_values fails
    return model


@pytest.fixture(scope="module")
def tokenizer(model_name: str) -> PreTrainedTokenizerBase:
    try:
        tokenizer = AutoTokenizer.from_pretrained(model_name)
    except Exception:
        # hf-internal-testing/tiny-random-MistralForCausalLM's tokenizer_config.json has
        # a field, tokenizer_file, which is hard-coded to some specific machine
        # Find it locally
        local_path = hf_hub.try_to_load_from_cache(model_name, "tokenizer.json")
        tokenizer = AutoTokenizer.from_pretrained(model_name, tokenizer_file=local_path)
    # Set attributes to values that would break CAPPr, if not for the context managers
    tokenizer.padding_side = "left"  # mismatched logits content b/c of position IDs
    if hasattr(tokenizer, "add_eos_token"):
        setattr(tokenizer, "add_eos_token", True)  # mismatched logits shape
    return tokenizer


@pytest.fixture(scope="module")
def model_and_tokenizer(
    model, tokenizer
) -> tuple[ModelForCausalLM, PreTrainedTokenizerBase]:
    return model, tokenizer


@pytest.fixture(scope="module")
def atol() -> float:
    # Reading through some transformers tests, it looks like 1e-3 is considered
    # close-enough for hidden states. See, e.g.,
    # https://github.com/huggingface/transformers/blob/897a826d830e8b1e03eb482b165b5d88a7a08d5f/tests/models/gpt2/test_modeling_gpt2.py#L252
    return 1e-4


########################################################################################
#################################### One-off tests #####################################
########################################################################################


def test_set_up_model(model: ModelForCausalLM):
    # Grab old attribute values
    attribute_to_old_value = {
        "training": model.training,
        **{i: module.training for i, module in enumerate(model.children())},
    }
    config_attribute_to_old_value = {
        "return_dict": model.config.return_dict,
        "use_cache": getattr(model.config, "use_cache", None),
    }
    assert torch.is_grad_enabled()

    # Enter the context
    with hf._utils.set_up_model(model):
        assert not torch.is_grad_enabled()
        assert not model.training
        if hasattr(model, "config"):
            # If the model doesn't have these attributes, then we'll see the error loud
            # and clear later in the tests. So default to the expected values
            assert getattr(model.config, "return_dict", True)
            assert getattr(model.config, "use_cache", True)

    # Exit the context. No attributes should have changed
    assert torch.is_grad_enabled()
    assert model.training == attribute_to_old_value["training"]
    for i, module in enumerate(model.children()):
        assert module.training == attribute_to_old_value[i]

    for attribute, old_value in config_attribute_to_old_value.items():
        assert getattr(model.config, attribute, None) == old_value


def test_set_up_tokenizer(tokenizer: PreTrainedTokenizerBase):
    # Grab old attribute values
    attributes = [
        "padding_side",
        "pad_token",
        "pad_token_id",
        "add_eos_token",
        "special_tokens_map",
    ]
    attribute_to_old_value = {
        attribute: getattr(tokenizer, attribute, None)
        # None is for tokenizers which don't have an add_eos_token attribute
        for attribute in attributes
    }

    # Enter the context
    with hf._utils.set_up_tokenizer(tokenizer):
        assert tokenizer.padding_side == "right"
        assert tokenizer.pad_token is not None
        assert tokenizer.pad_token_id is not None
        assert not getattr(tokenizer, "add_eos_token", False)

    # Exit the context. No attributes should have changed
    for attribute, old_value in attribute_to_old_value.items():
        assert getattr(tokenizer, attribute, None) == old_value


@pytest.mark.parametrize("texts", (["a b c", "d e", "f g h i"], ["slow reverb"]))
@pytest.mark.parametrize("batch_size", (1, 2))
def test__batched_model_call(texts, model, tokenizer, batch_size, atol):
    with hf._utils.set_up_tokenizer(tokenizer):
        encodings: BatchEncodingPT = tokenizer(texts, return_tensors="pt", padding=True)
    with hf._utils.set_up_model(model):
        out_correct: CausalLMOutput = model(**encodings)
    out_batched: CausalLMOutput = hf._utils._batched_model_call(
        batch_size=batch_size, model=model, **encodings
    )
    assert torch.allclose(out_correct.logits, out_batched.logits, atol=atol)


@pytest.mark.parametrize("module", (classify, classify_no_cache))
@pytest.mark.parametrize(
    "texts",
    (
        "lone string input",
        ["a b", "c d e"],
        ["a fistful", "of tokens", "for a few", "tokens more"],
    ),
)
@pytest.mark.parametrize("batch_size", (1, 2))
def test_token_logprobs(
    module: classify_module,
    texts,
    model_and_tokenizer,
    batch_size,
    end_of_prompt=" ",
    add_bos=False,
):
    """
    Tests that the model's token log probabilities are correct by testing against an
    unbatched and carefully, manually indexed result.
    """
    log_probs_texts_observed = module.token_logprobs(
        texts, model_and_tokenizer, end_of_prompt=end_of_prompt, batch_size=batch_size
    )

    model, tokenizer = model_and_tokenizer
    # Gather un-batched un-sliced log probs for the expected result
    is_str = isinstance(texts, str)
    texts = [texts] if is_str else texts
    if not hf._utils.does_tokenizer_need_prepended_space(tokenizer):
        end_of_prompt = ""
    log_probs_texts_from_unbatched = []
    input_ids_from_unbatched = []
    with hf._utils.dont_add_bos_token(tokenizer) if not add_bos else nullcontext():
        for text in texts:
            _logits, _encoding = hf._utils.logits_texts(
                [end_of_prompt + text], (model, tokenizer)
            )
            # grab first index b/c we only gave it 1 text
            log_probs_texts_from_unbatched.append(_logits[0].log_softmax(dim=1))
            input_ids_from_unbatched.append(_encoding["input_ids"][0])

    log_probs_texts_observed = (
        [log_probs_texts_observed] if is_str else log_probs_texts_observed
    )
    _test_content.token_logprobs(
        log_probs_texts_observed,
        log_probs_texts_from_unbatched,
        input_ids_from_unbatched,
    )


def test_cache_nested(model_and_tokenizer, atol):
    delim = " "
    if not hf._utils.does_tokenizer_need_prepended_space(model_and_tokenizer[1]):
        # for SentencePiece tokenizers like Llama's
        delim = ""

    logits = lambda *args, **kwargs: hf._utils.logits_texts(*args, **kwargs)[0]
    """
    Returns next-token logits for each token in an inputted text.
    """

    with classify.cache(model_and_tokenizer, "a") as cached_a:
        with classify.cache(cached_a, delim + "b c") as cached_a_b_c:
            with classify.cache(cached_a_b_c, delim + "d") as cached_a_b_c_d:
                logits1 = logits([delim + "e f"], cached_a_b_c_d)
                logits2 = logits([delim + "x"], cached_a_b_c_d)
            logits3 = logits([delim + "1 2 3"], cached_a_b_c)
        logits4 = logits([delim + "b c d"], cached_a)

    logits_correct = lambda texts, **kwargs: logits(
        texts, model_and_tokenizer, drop_bos_token=False
    )

    assert torch.allclose(logits1, logits_correct(["a b c d e f"]), atol=atol)
    assert torch.allclose(logits2, logits_correct(["a b c d x"]), atol=atol)
    assert torch.allclose(logits3, logits_correct(["a b c 1 2 3"]), atol=atol)
    assert torch.allclose(logits4, logits_correct(["a b c d"]), atol=atol)

    # Test clear_cache_on_exit
    device = model_and_tokenizer[0].device
    with pytest.raises(
        classify._CacheClearedError, match="This model is no longer usable."
    ):
        cached_a[0](
            input_ids=torch.tensor([[1]], device=device),
            attention_mask=torch.tensor([[1]], device=device),
        )

    with classify.cache(
        model_and_tokenizer, "a", clear_cache_on_exit=False
    ) as cached_a:
        with classify.cache(
            cached_a, delim + "b c", clear_cache_on_exit=False
        ) as cached_a_b_c:
            _ = logits(["whatever"], cached_a_b_c)
    assert cached_a_b_c[0]._cappr.past is not None
    assert hasattr(cached_a[0]._cappr, "past")

    # Test repr
    assert repr(cached_a[0]) == repr(cached_a[0]._cappr.model)


########################################################################################
################################### Helpers for tests ##################################
########################################################################################


def _test_encodings(
    logits_slow: torch.Tensor,
    encodings_slow: BatchEncodingPT,
    logits_fast: torch.Tensor,
    encodings_fast: BatchEncodingPT,
):
    """
    Tests that all objects have the expected shape, and that the encodings `offsets` are
    identical.
    """
    if logits_slow.shape[0] > logits_fast.shape[0] and logits_fast.shape[1] == 1:
        # Single-token optimization: this test doesn't apply b/c the optimization
        # doesn't repeat any data, unlike what's done in the slow/no-cache module
        return

    def _test_shapes(logits: torch.Tensor, encodings: BatchEncodingPT):
        assert encodings["input_ids"].shape == logits.shape[:2]  # 3rd dim is vocab
        assert encodings["input_ids"].shape == encodings["attention_mask"].shape
        assert encodings["input_ids"].shape[0] == encodings["offsets"].shape[0]

    _test_shapes(logits_slow, encodings_slow)
    _test_shapes(logits_fast, encodings_fast)

    # Test offsets. These should be exactly the same b/c they're the number of
    # of non-pad tokens in each prompt
    assert torch.equal(encodings_slow["offsets"], encodings_fast["offsets"])


def _test_logits(
    logits_slow: torch.Tensor,
    encodings_slow: BatchEncodingPT,
    logits_fast: torch.Tensor,
    encodings_fast: BatchEncodingPT,
    atol,
):
    """
    Tests that logits have identical shape, and that their non-pad token logits are
    numerically close.
    """
    if logits_slow.shape[0] > logits_fast.shape[0] and logits_fast.shape[1] == 1:
        # Single-token optimization: we only need to compare the last nonpad token's
        # logits for each prompt.
        num_completions = int(logits_slow.shape[0] / logits_fast.shape[0])
        logits_fast = logits_fast.repeat_interleave(num_completions, dim=0)
        last_nonpad_token_idxs = (encodings_slow["offsets"] - 1)[:, None, None]
        logits_slow_last_nonpad_token = logits_slow.take_along_dim(
            last_nonpad_token_idxs, dim=1
        )
        assert (
            logits_fast.shape == logits_slow_last_nonpad_token.shape
        )  # allclose doesn't check this
        assert torch.allclose(logits_fast, logits_slow_last_nonpad_token, atol=atol)
        return

    # Middle dimension (for the # of tokens) is different b/c logits_slow includes
    # prompt and completion tokens, while logits_fast only includes completion tokens.
    assert logits_slow.shape[2] == logits_fast.shape[2]  # vocab size

    # Test logits at every *non-pad* token
    completion_token_idxs = [
        list(range(num_completion_tokens))
        for num_completion_tokens in encodings_fast["attention_mask"].sum(dim=1)
    ]
    for text_idx in range(logits_slow.shape[0]):
        offset = encodings_fast["offsets"][text_idx].item() - 1
        # number of non-pad prompt tokens - 1 (!) b/c in the fast version we
        # included the last non-pad prompt token
        for completion_token_idx in completion_token_idxs[text_idx]:
            assert torch.allclose(
                logits_fast[text_idx, completion_token_idx],
                logits_slow[text_idx, offset + completion_token_idx],
                atol=atol,
            )


########################################################################################
####################################### Tests ##########################################
########################################################################################


class Modules:
    @property
    def module(self):
        return classify

    @property
    def module_correct(self):
        return classify_no_cache


class TestPromptsCompletions(Modules, _base.TestPromptsCompletions):
    def test__logits_completions_given_prompts(
        self, model, tokenizer, prompts, completions, atol
    ):
        # Test logits for better debuggability
        # For this helper function, prompts can't be a single string
        if isinstance(prompts, str):
            return
        slow_out = classify_no_cache._logits_completions_given_prompts(
            model, tokenizer, prompts, completions
        )
        fast_out = classify._logits_completions_given_prompts(
            model, tokenizer, prompts, completions
        )
        _test_encodings(*slow_out, *fast_out)
        _test_logits(*slow_out, *fast_out, atol)

    @pytest.mark.parametrize("batch_size", (1, 2))
    @pytest.mark.parametrize("batch_size_completions", (None, 1))
    def test_log_probs_conditional(
        self,
        prompts,
        completions,
        model_and_tokenizer,
        batch_size,
        batch_size_completions,
    ):
        super().test_log_probs_conditional(
            prompts,
            completions,
            model_and_tokenizer,
            batch_size=batch_size,
            batch_size_completions=batch_size_completions,
        )

    def test_predict_proba(
        self,
        prompts,
        completions,
        model_and_tokenizer,
        prior,
        discount_completions,
        normalize,
    ):
        super().test_predict_proba(
            prompts,
            completions,
            model_and_tokenizer,
            prior=prior,
            discount_completions=discount_completions,
            normalize=normalize,
        )

    def test_predict(self, prompts, completions, model_and_tokenizer):
        super().test_predict(prompts, completions, model_and_tokenizer)


class TestExamples(Modules, _base.TestExamples):
    def test__logits_completions_given_prompts_examples(
        self, model, tokenizer, examples, atol
    ):
        # Test logits for better debuggability
        # For this helper function, examples can't be an Example
        if isinstance(examples, Example):
            return
        slow_out = classify_no_cache._logits_completions_given_prompts_examples(
            model, tokenizer, examples
        )
        fast_out = classify._logits_completions_given_prompts_examples(
            model, tokenizer, examples
        )
        _test_encodings(*slow_out, *fast_out)
        _test_logits(*slow_out, *fast_out, atol)

    @pytest.mark.parametrize("batch_size", (1, 2))
    def test_log_probs_conditional_examples(
        self, examples: Example | Sequence[Example], model_and_tokenizer, batch_size
    ):
        super().test_log_probs_conditional_examples(
            examples, model_and_tokenizer, batch_size=batch_size
        )

    def test_predict_proba_examples(
        self, examples: Example | Sequence[Example], model_and_tokenizer
    ):
        super().test_predict_proba_examples(examples, model_and_tokenizer)

    def test_predict_examples(
        self, examples: Example | Sequence[Example], model_and_tokenizer
    ):
        super().test_predict_examples(examples, model_and_tokenizer)


@pytest.mark.parametrize("batch_size", (1, 2))
class TestCache(Modules, _base.TestCache):
    def test_cache(
        self,
        prompt_prefix: str,
        prompts: list[str],
        completions: list[str],
        model_and_tokenizer,
        batch_size,
    ):
        super().test_cache(
            prompt_prefix,
            prompts,
            completions,
            model_and_tokenizer,
            batch_size=batch_size,
        )

    def test_cache_model(
        self,
        prompt_prefix: str,
        prompts: list[str],
        completions: list[str],
        model_and_tokenizer,
        batch_size,
    ):
        super().test_cache_model(
            prompt_prefix,
            prompts,
            completions,
            model_and_tokenizer,
            batch_size=batch_size,
        )



from __future__ import annotations
from typing import Optional, Union

from IPython.display import display
import pandas as pd


def display_df(
    df: pd.DataFrame,
    columns: Optional[list[str]] = None,
    num_rows: Union[int, None] = 3,
):
    """
    Displays `df.head(num_rows)[columns]` without truncating columns. If
    possible, render any newlines.
    """
    if columns is None:
        columns = df.columns
    if num_rows is None:
        num_rows = len(df)
    df_head_styled = df.head(num_rows)[columns].style
    with pd.option_context("max_colwidth", None):
        ## I'm not sure why try-except doesn't work w/ display(), so instead
        ## check the necessary uniqueness condition before running it
        if df.index.is_unique:
            display(
                df_head_styled.set_properties(
                    **{"text-align": "left", "white-space": "pre-wrap"}
                )
            )
        else:
            ## `Styler.apply` and `.applymap` are not compatible with non-unique
            ## index or columns
            display(df_head_styled)


def remove_suffix(string: str, suffix: str):
    if string.endswith(suffix):
        return string[: -len(suffix)]
    return string


def remove_prefix(string: str, prefix: str) -> str:
    if string.startswith(prefix):
        return string[len(prefix) :]
    return string



"""
Completion After Prompt Probability. Make your LLM make a choice

https://cappr.readthedocs.io/
"""

__version__ = "0.9.1"

from . import utils  # noqa: F401
from ._example import Example  # noqa: F401

try:
    from . import openai  # noqa: F401
except ModuleNotFoundError:  # pragma: no cover
    pass

try:
    from . import huggingface  # noqa: F401
except ModuleNotFoundError:  # pragma: no cover
    pass

try:
    from . import llama_cpp  # noqa: F401
except ModuleNotFoundError:  # pragma: no cover
    pass



"""
YouTils
"""

from __future__ import annotations
from contextlib import contextmanager, ExitStack, nullcontext
from functools import lru_cache
from typing import Mapping, Sequence

import torch
from transformers import PreTrainedModel, PreTrainedTokenizerBase
from transformers.modeling_outputs import CausalLMOutput

from cappr.utils import _check
from cappr.utils.classify import _setattr


BatchEncodingPT = Mapping[str, torch.Tensor]
"""
Type of the output of `tokenizer(texts, return_tensors="pt", ...)`
"""
# transformers.BatchEncoding doesn't annotate values as tensors b/c tokenizers can
# return other objects. In this package, tokenizers will always return PyTorch tensors


ModelForCausalLM = PreTrainedModel
"""
A pretrained model with the same inference interface as a model loaded via
`transformers.AutoModelForCausalLM.from_pretrained`
"""


########################################################################################
# Set up the model and tokenizer to enable correct, batched inference. Use context
# managers b/c a user's settings shouldn't be modified when CAPPr is done. It's
# conceivable that someone uses CAPPr as part of a larger system where the model and
# tokenizer needs to be configured differently
########################################################################################


@contextmanager
def _eval_mode(model: ModelForCausalLM):
    """
    In this context, the model is set in eval mode.
    """
    is_training = model.training
    try:
        model.eval()
        yield
    finally:
        model.train(is_training)


@contextmanager
def _no_grad(model: ModelForCausalLM):  # model given to keep interface the same
    """
    In this context, gradients are not computed.
    """
    with torch.no_grad():
        yield


# Some models don't perfectly implement the HF model call interface. In particular,
# they're missing the return_dict and use_cache kwargs. They're instead in the model
# config. I see that as a more extensible design anyway.


@contextmanager
def _return_dict(model: ModelForCausalLM):
    """
    In this context, the model returns a dataclass when it's called.
    """
    with (
        _setattr(model.config, "return_dict", True)
        if hasattr(model, "config")
        else nullcontext()
    ):  # null b/c just try model(...).logits when needed
        yield


@contextmanager
def _use_cache(model: ModelForCausalLM):
    """
    In this context, the model output includes a `past_key_values` attribute.
    """
    with (
        _setattr(model.config, "use_cache", True)
        if hasattr(model, "config")
        else nullcontext()
    ):  # null b/c just try model(...).past_key_values when needed
        yield


@contextmanager
def _pad_on_right(tokenizer: PreTrainedTokenizerBase):
    """
    In this context, the pad token is set (if it's not set already) to the EOS token so
    that batch inference is possible. These get masked out. The padding side is set to
    right. Left-padding would alter position IDs for non-pad tokens, which makes things
    a bit more confusing.
    """
    pad_token_id = tokenizer.pad_token_id
    padding_side = tokenizer.padding_side
    try:
        if pad_token_id is None:
            tokenizer.pad_token_id = tokenizer.eos_token_id
            # PreTrainedTokenizerBase is smart about setting auxiliary attributes, e.g.,
            # tokenizer.special_tokens_map is updated after tokenizer.pad_token_id = ...
        tokenizer.padding_side = "right"
        yield
    finally:
        tokenizer.pad_token_id = pad_token_id
        tokenizer.padding_side = padding_side


@contextmanager
def dont_add_eos_token(tokenizer: PreTrainedTokenizerBase):
    """
    In this context, don't add an end-of-sentence token.
    """
    with _setattr(tokenizer, "add_eos_token", False):
        yield


@contextmanager
def _combine_context_managers(context_managers, obj):
    with ExitStack() as stack:
        [
            stack.enter_context(context_manager(obj))
            for context_manager in context_managers
        ]
        yield


@contextmanager
def set_up_model(
    model: ModelForCausalLM,
    context_managers=(_eval_mode, _no_grad, _return_dict, _use_cache),
):
    with _combine_context_managers(context_managers, model):
        yield


@contextmanager
def set_up_tokenizer(
    tokenizer: PreTrainedTokenizerBase,
    context_managers=(_pad_on_right, dont_add_eos_token),
):
    with _combine_context_managers(context_managers, tokenizer):
        yield


@contextmanager
def dont_add_bos_token(tokenizer: PreTrainedTokenizerBase):
    """
    In this context, don't add a beginning-of-sentence token.
    """
    with _setattr(tokenizer, "add_bos_token", False):
        yield


########################################################################################
####################### Handle SentencePiece vs BPE tokenization #######################
########################################################################################


@lru_cache()
def does_tokenizer_need_prepended_space(
    tokenizer: PreTrainedTokenizerBase,
) -> bool:
    with dont_add_eos_token(tokenizer):
        tokenize = lambda text: tokenizer(text)["input_ids"]
        bos_token_id = getattr(tokenizer, "bos_token_id", None)
        return _check.does_tokenizer_need_prepended_space(tokenize, bos_token_id)


########################################################################################
##################################### Logits stuff #####################################
########################################################################################


def _batched_model_call(
    batch_size: int,
    model: ModelForCausalLM,
    input_ids: torch.Tensor,
    attention_mask: torch.Tensor,
) -> CausalLMOutput:
    input_ids = torch.split(input_ids, batch_size)
    attention_mask = torch.split(attention_mask, batch_size)
    outs: list[CausalLMOutput] = []
    num_batches = len(input_ids)
    with set_up_model(model):
        for batch_idx in range(num_batches):
            out: CausalLMOutput = model(
                input_ids=input_ids[batch_idx],
                attention_mask=attention_mask[batch_idx],
            )
            outs.append(out)
    logits = torch.cat([out.logits for out in outs], dim=0)
    return CausalLMOutput(logits=logits)


def logits_texts(
    texts: Sequence[str],
    model_and_tokenizer: tuple[ModelForCausalLM, PreTrainedTokenizerBase],
    drop_bos_token: bool = True,
    batch_size: int | None = None,
) -> tuple[torch.Tensor, BatchEncodingPT]:
    """
    Basically::

        return model(**tokenizer(texts)).logits, tokenizer(texts)

    The kwargs are set for CAPPr's convenience.
    """
    model, tokenizer = model_and_tokenizer
    with set_up_tokenizer(tokenizer):
        encodings: BatchEncodingPT = tokenizer(
            texts, return_tensors="pt", padding=True
        ).to(model.device)
    if batch_size is not None:
        out = _batched_model_call(batch_size, model, **encodings)
    else:
        with set_up_model(model):
            out: CausalLMOutput = model(**encodings)
    if drop_bos_token and getattr(tokenizer, "add_bos_token", False):
        # Drop the first/bos token after we're done encoding so that the shape is
        # consistent w/ other tokenizers. For CAPPr, we'll never be interested in
        # Pr(token | <bos>). We're only interested in completion tokens
        logits = out.logits[:, 1:, :]
        encodings = {key: value[:, 1:] for key, value in encodings.items()}
    else:
        logits = out.logits
    return logits, encodings


def logits_to_log_probs(
    logits: torch.Tensor,
    input_ids: torch.Tensor,
    input_ids_start_idx: int | None = None,
    logits_end_idx: int | None = None,
):
    """
    Log-softmax and then slice out input IDs to get token log-probabilities.
    """
    # logits.shape is (# texts, max # tokens in texts, vocab size)
    log_probs = logits.log_softmax(dim=-1)

    # Only keep the log-prob from the vocab dimension whose index is is the next token's
    # input ID
    # input_ids.shape is (# texts, max # tokens in texts)
    return (
        log_probs[:, :logits_end_idx, :]
        .take_along_dim(input_ids[:, input_ids_start_idx:, None], dim=-1)
        .squeeze(-1)
    )



"""
Unit and integration tests for `cappr.llama_cpp.classify`. Works by checking that its
functions' outputs are numerically close to those from
`cappr.llama_cpp._classify_no_cache`.
"""

from __future__ import annotations
from dataclasses import dataclass
import os
import sys
from typing import Sequence

from huggingface_hub import hf_hub_download
from llama_cpp import Llama
import numpy as np
import pytest
import torch

from cappr import Example
from cappr.llama_cpp import _utils, classify, _classify_no_cache

# sys hack to import from parent
sys.path.insert(1, os.path.join(sys.path[0], ".."))
import _base
import _test_content


_MODELS_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "fixtures", "models"
)


########################################################################################
###################################### Fixtures ########################################
########################################################################################


@dataclass(frozen=True)
class _HFHubModel:
    repo_id: str
    filename: str
    does_tokenizer_need_prepended_space: bool


@pytest.fixture(
    scope="module",
    params=[
        _HFHubModel(
            repo_id="aladar/tiny-random-BloomForCausalLM-GGUF",
            filename="tiny-random-BloomForCausalLM.gguf",
            does_tokenizer_need_prepended_space=True,  # BPE
        ),
        _HFHubModel(
            repo_id="aladar/tiny-random-LlamaForCausalLM-GGUF",
            filename="tiny-random-LlamaForCausalLM.gguf",
            does_tokenizer_need_prepended_space=False,  # SentencePiece
        ),
    ],
)
def model_and_does_tokenizer_need_prepended_space(
    request: pytest.FixtureRequest,
) -> tuple[Llama, bool]:
    hf_hub_model: _HFHubModel = request.param
    model_path = hf_hub_download(
        hf_hub_model.repo_id, hf_hub_model.filename, local_dir=_MODELS_DIR
    )
    model = Llama(model_path, verbose=False)
    return model, hf_hub_model.does_tokenizer_need_prepended_space


@pytest.fixture(scope="module")
def model(model_and_does_tokenizer_need_prepended_space: tuple[Llama, bool]) -> Llama:
    model, _ = model_and_does_tokenizer_need_prepended_space
    # Correctness should not be affected by a user's previous actions, which get cached
    model.eval(model.tokenize("a b c".encode()))
    return model


@pytest.fixture(scope="module")
def atol() -> float:
    # Reading through some transformers tests, it looks like 1e-3 is considered
    # close-enough for hidden states. See, e.g.,
    # https://github.com/huggingface/transformers/blob/897a826d830e8b1e03eb482b165b5d88a7a08d5f/tests/models/gpt2/test_modeling_gpt2.py#L252
    return 1e-4


########################################################################################
#################################### One-off tests #####################################
########################################################################################


def test__does_tokenizer_need_prepended_space(
    model_and_does_tokenizer_need_prepended_space,
):
    """
    Explicitly test this b/c w/o it, headaches are possible.
    """
    (
        model,
        does_tokenizer_need_prepended_space_expected,
    ) = model_and_does_tokenizer_need_prepended_space
    assert (
        _utils.does_tokenizer_need_prepended_space(model)
        == does_tokenizer_need_prepended_space_expected
    )
    # If it's cached, let's double check that we still get the expected result
    assert (
        _utils.does_tokenizer_need_prepended_space(model)
        == does_tokenizer_need_prepended_space_expected
    )


def test_set_up_model(model: Llama):
    assert not model.context_params.logits_all
    with _utils.set_up_model(model):
        assert model.context_params.logits_all
    assert not model.context_params.logits_all


@pytest.mark.parametrize("shape_and_dim", [((10,), 0), ((3, 10), 1)])
def test_log_softmax(
    shape_and_dim: tuple[tuple[int] | tuple[int, int], int], atol: float
):
    shape, dim = shape_and_dim
    data: np.ndarray = np.random.randn(*shape)
    log_probs_observed = _utils.log_softmax(data, dim=dim)
    log_probs_expected = torch.tensor(data).log_softmax(dim=dim).numpy()
    assert np.allclose(log_probs_observed, log_probs_expected, atol=atol)


@pytest.mark.parametrize(
    "texts",
    (
        "lone string input",
        ["a b", "c d e"],
        ["a fistful", "of tokens", "for a few", "tokens more"],
    ),
)
def test_token_logprobs(texts: Sequence[str], model: Llama, end_of_prompt=""):
    """
    Tests that the model's token log probabilities are correct by testing against a
    carefully, manually indexed result.
    """
    log_probs_texts_observed = classify.token_logprobs(
        texts, model, add_bos=True, end_of_prompt=end_of_prompt
    )

    # Gather un-batched un-sliced log probs for the expected result
    is_str = isinstance(texts, str)
    texts = [texts] if is_str else texts
    if not _utils.does_tokenizer_need_prepended_space(model):
        end_of_prompt = ""
    log_probs_texts_from_unbatched = []
    input_ids_from_unbatched = []
    with _utils.set_up_model(model):
        for text in texts:
            input_ids = model.tokenize((end_of_prompt + text).encode(), add_bos=True)
            model.reset()
            model.eval(input_ids)
            log_probs_texts_from_unbatched.append(
                _utils.log_softmax(np.array(model.eval_logits))
            )
            input_ids_from_unbatched.append(input_ids)
    model.reset()

    log_probs_texts_observed = (
        [log_probs_texts_observed] if is_str else log_probs_texts_observed
    )
    _test_content.token_logprobs(
        log_probs_texts_observed,
        log_probs_texts_from_unbatched,
        input_ids_from_unbatched,
    )


########################################################################################
####################################### Tests ##########################################
########################################################################################


class Modules:
    @property
    def module(self):
        return classify

    @property
    def module_correct(self):
        return _classify_no_cache


class TestPromptsCompletions(Modules, _base.TestPromptsCompletions):
    def test_log_probs_conditional(self, prompts, completions, model):
        super().test_log_probs_conditional(prompts, completions, model)

    def test_predict_proba(
        self,
        prompts,
        completions,
        model,
        prior,
        discount_completions,
        normalize,
    ):
        super().test_predict_proba(
            prompts,
            completions,
            model,
            prior=prior,
            discount_completions=discount_completions,
            normalize=normalize,
        )

    def test_predict(self, prompts, completions, model):
        super().test_predict(prompts, completions, model)


class TestExamples(Modules, _base.TestExamples):
    def test_log_probs_conditional_examples(
        self, examples: Example | Sequence[Example], model
    ):
        super().test_log_probs_conditional_examples(examples, model)

    def test_predict_proba_examples(self, examples: Example | Sequence[Example], model):
        super().test_predict_proba_examples(examples, model)

    def test_predict_examples(self, examples: Example | Sequence[Example], model):
        super().test_predict_examples(examples, model)


class TestCache(Modules, _base.TestCache):
    def _test_log_probs_conditional(self, *args, **kwargs):
        model: Llama = args[-1]
        model.reset()  # reset cache before evaluating correct module
        super()._test_log_probs_conditional(*args, **kwargs)

    def _test_log_probs_conditional_examples(self, *args, **kwargs):
        model: Llama = args[-1]
        model.reset()  # reset cache before evaluating correct module
        super()._test_log_probs_conditional_examples(*args, **kwargs)

    def test_cache(
        self, prompt_prefix: str, prompts: list[str], completions: list[str], model
    ):
        super().test_cache(
            prompt_prefix, prompts, completions, model, reset_model=False
        )

    def test_cache_model(
        self, prompt_prefix: str, prompts: list[str], completions: list[str], model
    ):
        super().test_cache_model(
            prompt_prefix, prompts, completions, model, reset_model=False
        )



"""
Shared input checks. These check conditions that would cause silent, difficult-to-debug,
or expensive errors.

For example:
    - inputting an unordered iterable will cause an output array of predicted
      probabilities to be meaningless
    - inputting an empty object can cause an obscure index error in some downstream
      model/tokenization functions
    - inputting an incorrectly structured prior will cause all of the model's compute to
      be a waste.
"""

from __future__ import annotations
from typing import Callable, Literal, Sequence

import numpy as np
import numpy.typing as npt


def _is_reversible(object) -> bool:
    # Returns True for:
    # - list, tuple, dict keys, dict values
    # - numpy array, torch Tensor
    # - pandas and polars Series
    # - str (but other places in this package filter those out before getting here)
    # Returns False for:
    # - set
    try:
        reversed(object)  # often a generator, so checking this is often free
    except TypeError:
        return False
    else:
        return True


def ordered(object: Sequence, variable_name: str):
    """
    Raises a `TypeError` is `object` is not a sequence.
    """
    # Just want [x for x in object] to be meaningful and deterministic
    # isinstance(object, Sequence) is too restrictive. Sequence requires that
    # object.__class__ additionally implements index and count. See:
    # https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes
    if not _is_reversible(object):
        raise TypeError(
            f"{variable_name} must be an ordered collection. Consider converting it to "
            "a list or tuple."
        )


def nonempty(object: Sequence, variable_name: str):
    """
    Raises a `ValueError` if `object` is empty.
    """
    if len(object) == 0:
        raise ValueError(f"{variable_name} must be non-empty.")


def nonempty_and_ordered(object: Sequence, variable_name: str):
    """
    Raises an error if `object` is not nonempty and ordered.
    """
    nonempty(object, variable_name)
    ordered(object, variable_name)


def completions(completions: Sequence[str]):
    """
    Raises an error if `completions` is not a nonempty, ordered, non-string, or if it
    contains an empty string.
    """
    nonempty_and_ordered(completions, variable_name="completions")
    if isinstance(completions, str):
        raise TypeError(
            "completions cannot be a string. It must be a sequence of strings. If you "
            "intend on inputting a single completion to estimate its probability, wrap "
            "it in a list or tuple (and set normalize=False in a predict_proba call)."
        )
    idxs_empty_completions = [
        i for i, completion in enumerate(completions) if not completion
    ]
    if len(idxs_empty_completions) == len(completions):
        raise ValueError(
            "All completions are empty. Expected all completions to be non-empty "
            "strings."
        )
    elif idxs_empty_completions:
        raise ValueError(
            f"completions {idxs_empty_completions} are empty. Expected all completions "
            "to be non-empty strings."
        )


def end_of_prompt(end_of_prompt: Literal[" ", ""]):
    """
    Raises an error if `end_of_prompt` is not a whitespace or empty string.
    """
    msg = 'end_of_prompt must be a whitespace " " or empty string "".'
    if not isinstance(end_of_prompt, str):
        raise TypeError(msg)
    if end_of_prompt not in {" ", ""}:
        raise ValueError(msg)


def prior(
    prior: Sequence[float] | None, expected_length: int
) -> npt.NDArray[np.floating] | None:
    """
    Return back `prior` if it's None, or return it as a numpy array if it's not None.

    Raises an error if `prior is not None` and isn't a probability distribution over
    `expected_length` categories.
    """
    if prior is None:  # it's a uniform prior, no need to check anything
        return None
    # We do the prior * likelihood computation after making expensive model calls. Don't
    # want that multiplication to fail b/c that'd be a complete waste of model compute
    nonempty_and_ordered(prior, variable_name="prior")
    if len(np.shape(prior)) != 1:
        raise ValueError("prior must be 1-D.")
    prior_arr = np.array(prior, dtype=float)  # try casting to float
    if not np.all(prior_arr >= 0):
        raise ValueError("prior must contain probabilities between 0 and 1.")
    if not np.all(prior_arr <= 1):
        raise ValueError("prior must contain probabilities between 0 and 1.")
    if not np.isclose(prior_arr.sum(), 1):
        raise ValueError("prior must sum to 1.")
    if len(prior) != expected_length:
        raise ValueError(
            f"Expected prior to have length {expected_length} (the number of "
            f"completions), got {len(prior)}."
        )
    return prior_arr


def normalize(normalize: bool, completions: Sequence[str]):
    """
    Raises a `ValueError` if `normalize and len(completions) == 1`.
    """
    if normalize and len(completions) == 1:
        raise ValueError(
            "Setting normalize=True when there's only 1 completion causes the "
            "probability to trivially be 1. Did you mean to set normalize=False, or "
            "did you mean to include more completions?"
        )


def does_tokenizer_need_prepended_space(
    tokenize: Callable[[str], list[int]], bos_token_id: int | None
) -> bool:
    """
    Returns `True` if the `tokenize` function needs to prepend a whitespace for
    subsequent model calls to be correct, else `False`. BPE tokenizers need it, while
    SentencePiece tokenizers do not.
    """

    def remove_bos(tokens: list[int]) -> list[int]:
        if bos_token_id is None:
            return tokens
        if tokens[0] == bos_token_id:
            return tokens[1:]
        return tokens

    tokens = tokenize("a b")
    tokens_concat = tokenize("a") + remove_bos(tokenize("b"))
    if tokens != tokens_concat:
        tokens_concat_correct = tokenize("a") + remove_bos(tokenize(" b"))
        if tokens != tokens_concat_correct:
            raise ValueError(
                "This tokenizer is weird. Perhaps it's adding EOS tokens? Please raise "
                "this as an issue so that I can investigate: "
                "https://github.com/kddubey/cappr/issues"
            )  # pragma: no cover
        return True
    return False



"""
Mirror. For testing purposes only. Is strictly slower
"""

from __future__ import annotations
from functools import partial
from typing import Literal, Sequence

from llama_cpp import Llama
import numpy as np
import numpy.typing as npt

from cappr.utils import _no_cache, classify
from cappr import Example
from cappr.llama_cpp.classify import token_logprobs


def _tokenize_completions(model: Llama, texts: Sequence[str]) -> list[list[int]]:
    return [model.tokenize(text.encode(), add_bos=False) for text in texts]


@classify._log_probs_conditional
def log_probs_conditional(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    **kwargs,
) -> list[list[float]] | list[list[list[float]]]:
    return _no_cache.log_probs_conditional(
        prompts,
        completions,
        end_of_prompt,
        token_logprobs,
        partial(_tokenize_completions, model),
        model.token_bos(),
        model,
        add_bos=True,
    )


@classify._log_probs_conditional_examples
def log_probs_conditional_examples(
    examples: Example | Sequence[Example],
    model: Llama,
    **kwargs,
) -> list[list[float]] | list[list[list[float]]]:
    return _no_cache.log_probs_conditional_examples(
        examples,
        token_logprobs,
        partial(_tokenize_completions, model),
        model.token_bos(),
        model,
        add_bos=True,
    )


@classify._predict_proba
def predict_proba(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    prior: Sequence[float] | None = None,
    normalize: bool = True,
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
) -> npt.NDArray[np.floating]:
    return log_probs_conditional(**locals())


@classify._predict_proba_examples
def predict_proba_examples(
    examples: Example | Sequence[Example],
    model: Llama,
) -> npt.NDArray[np.floating] | list[npt.NDArray[np.floating]]:
    return log_probs_conditional_examples(**locals())


@classify._predict
def predict(
    prompts: str | Sequence[str],
    completions: Sequence[str],
    model: Llama,
    end_of_prompt: Literal[" ", ""] = " ",
    prior: Sequence[float] | None = None,
    discount_completions: float = 0.0,
    log_marg_probs_completions: Sequence[Sequence[float]] | None = None,
) -> str | list[str]:
    return predict_proba(**locals())


@classify._predict_examples
def predict_examples(
    examples: Example | Sequence[Example],
    model: Llama,
) -> str | list[str]:
    return predict_proba_examples(**locals())



"""
Install the Hugging Face requirements to use this module::

    pip install "cappr[hf]"

See this section of the documentation:

https://cappr.readthedocs.io/en/latest/select_a_language_model.html#hugging-face
"""

from . import _utils, classify, classify_no_cache

__all__ = ["_utils", "classify", "classify_no_cache"]
